#!/usr/bin/perl

use Switch;
use List::Util qw(first);

my $use_limboole = 1;

sub println {
    local $\ = "\n";
    print @_;
}

sub trim {
	@_ == 1 || die("ERROR: too few/many arguments for trim()\n\tusage: trim(str)\n");
	
	my $str = $_[0];
	$str =~ s/^\s+|\s+$//g;
	
	return $str;
}

sub process_universal {
	@_ == 1 || die("ERROR: too few/many arguments for process_universal()\n\tusage: process_universal(exp)\n");
	my $exp = trim substr($_[0], 1);

	push(@uvars, split(/ +/,$exp));
}

sub process_existential {
	@_ == 1 || die("ERROR: too few/many arguments for process_existential()\n\tusage: process_existential(exp)\n");
	my $exp = trim substr($_[0], 1);
	
	$exp =~ s/[\{\}]//g;
	
	my @vars = split(/ +/,$exp);
	
	push(@evars, pop(@vars));
	push(@deps, [ @vars ]);
}

sub flatten_deps {
	my $varInd = $_[0];
	my @newAllDeps = ();	
	
	foreach my $dep (@{$deps[$varInd]}) {
		if ($dep ~~ @uvars) {
			if (!($dep ~~ @newAllDeps)) { push(@newAllDeps, $dep); }
			next;
		}
		
		$varInd = first { $evars[$_] eq $dep } 0..$#evars;
		
		my @newDeps = flatten_deps($varInd);
		
		foreach my $newDep (@newDeps) {
			if ($newDep ~~ @newAllDeps) { next; }
			push(@newAllDeps, $newDep);
		}
	}
	
	return @newAllDeps;
}

sub process_formula {
	@_ == 1 || die("ERROR: too few/many arguments for process_formula()\n\tusage: process_formula(exp)\n");
	my $exp = $_[0];

	if ($use_limboole) { $exp =~ s/<-->/<->/g; }
	else { $exp =~ s/<-->/=/g; }
	$exp =~ s/\\\//|/g;
	$exp =~ s/\/\\/&/g;
	if (!$use_limboole) { $exp =~ s/~/!/g; }
	
	if ($use_limboole) { 
		system("echo \"$exp\" | limboole/limboole -d -s > out.dimacs") == 0
			|| die "executing limboole did not succedd: $!";
	}
	else {
		system("echo \"$exp\" | bool2cnf/bool2cnf -s > out.dimacs") == 0
			|| die "executing bool2cnf did not succedd: $!";
	}
		
	open(DIMACS, "<", "out.dimacs")
		|| die "can't open out.dimacs: $!";

	my $in_header = 1;
	while (<DIMACS>) {
		switch (substr($_, 0, 1)) {
			case "c" {
			}
			case "p" {
				$in_header = 0;
				goto END_LOOP;
			}
			else {
				goto END_LOOP;
			}
		}
		
		if ($in_header) {
			my $id, $var;
			if ($use_limboole) { ($id, $var) = split(/ +/, trim(substr($_, 2))); }
			else { ($var, $id) = split(/ +/, trim(substr($_, 6))); }
			foreach my $uvar (@uvars) {
				if ($uvar eq $var) {
					$uvar = $id;
					foreach my $dep_list (@deps) {
						foreach my $dep (@$dep_list) {
							if ($dep eq $var) { $dep = $id; }
						}
					}
				}
			}
			foreach my $evar (@evars) {
				if ($evar eq $var) { $evar = $id; }
			}
		}
		else {
			if ($use_limboole) {
				my @tokens = split(/ +/, trim(substr($_, 2)));
				push(@evars, shift(@tokens));
				push(@deps, [ @tokens ]);
			}
		}
		
		END_LOOP:
	}
	
	for my $i (0 .. $#evars) {
		if ($evars[$i] =~ /^\d+$/) {
			$deps[$i] = [ flatten_deps($i) ];
		}
	}
	
	
	seek DIMACS, 0, 0;

	while (<DIMACS>) {
		switch (substr($_, 0, 1)) {
			case "c" {
			}
			case "p" {
				print $_;
				(my $cntVars) = split(/ +/, substr($_, 6));

				for my $i (1 .. $cntVars) { $var_used[$id] = 0; }

				print "a ";
				foreach my $uvar (@uvars) {
					if ($uvar =~ /^\d+$/) {
						print "$uvar ";
						$var_used[$uvar] = 1;
					}
				}
				println "0";
				
				for my $i (0 .. $#evars) {
					if ($evars[$i] =~ /^\d+$/) { 
						print "d $evars[$i] ";
						$var_used[$evars[$i]] = 1;
						foreach my $dep (@{$deps[$i]}) {
							if ($dep =~ /^\d+$/) { print "$dep "; }
						}
						println "0";
					}
				}
			}
			else {
				print $_;
			}
		}
	}
	
	close DIMACS;
	
	unlink "out.dimacs";
}


#======= MAIN FUNCTION


$,=' ';

while(<>) {
while (true) {
	($exp, $_) = split(":", $_, 2);
	$exp = trim($exp);
	if ($exp eq "") { last; }
	
	switch (substr($exp, 0, 1)) {
		case "A" { process_universal($exp); }
		case "E" { process_existential($exp); }
		else { process_formula($exp); }
	}
}
}